#include "lenswidget.h"
#include <grale/realtype.h>
#include <grale/sislens.h>
#include <grale/compositelens.h>
#include <grale/constants.h>
#include <grale/sourceplane.h>
#include <grale/imageplane.h>
#include <grale/lensplane.h>
#include <grale/real2dlensplanefunction.h>
#include <grale/openclkernel.h>
#include <qimage.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qevent.h>
#include <qmessagebox.h>
#include <q3progressbar.h>
//Added by qt3to4:
#include <QMouseEvent>

#include <iostream> // TODO

#define MAXCIRCLES 10
//#define SUBDIV

using namespace grale;

LensWidget::LensWidget(QWidget *pParent, int width, int height, double angularWidth) : 
	Q3ScrollView(pParent, 0, Qt::WNoAutoErase|Qt::WPaintClever)
{
	resizeContents(width, height);
	m_pProgressBar = 0;

	m_width = width;
	m_height = height;
	m_angularWidth = angularWidth;
	
	m_mouseStartX = -1;
	m_mouseStartY = -1;

	m_pImagePlane = 0;

	m_pBackGroundImage = 0;
	m_pCachedView = 0;

	m_massScale = 1.0;

#ifdef GRALECONFIG_SUPPORT_OPENCL // TODO: add a command to do this
	m_pCLKernel = new grale::OpenCLKernel();

	if (!m_pCLKernel->init())
	{
		delete m_pCLKernel;
		m_pCLKernel = 0;
	}
#else
	m_pCLKernel = 0;
#endif // GRALECONFIG_SUPPORT_OPENCL
}


LensWidget::~LensWidget()
{
}

void LensWidget::drawContents(QPainter *pPainter, int clipx, int clipy, int clipw, int cliph)
{
	if (m_width != contentsWidth() || m_height != contentsHeight())
		resizeContents(m_width, m_height);

	QImage image;

	if (m_pCachedView == 0)
		m_pCachedView = new QImage(getLensedImage());
	
	image = *m_pCachedView;

	QPixmap bufferPixmap(image);
	QPainter dbPainter(&bufferPixmap);

#if 0
	for (int i = 0 ; i < m_circles.size() ; i++)
	{
		dbPainter.setPen(QPen(QColor(0,255,0), 2));
		
		int startX = m_circles[i].m_cx-m_circles[i].m_width/2;
		int startY = m_circles[i].m_cy-m_circles[i].m_width/2;

		dbPainter.drawEllipse(startX, startY, m_circles[i].m_width, m_circles[i].m_width);
	}
#endif

	if (m_mouseStartX != -1)
	{
		dbPainter.setPen(QPen(QColor(0,255,255), 2));
		
		int startX = m_mouseStartX-m_circleWidth/2;
		int startY = m_mouseStartY-m_circleWidth/2;

		dbPainter.drawEllipse(startX, startY, m_circleWidth, m_circleWidth);
		m_lastDrawnWidth = m_circleWidth;
	}

#if 0
	// Critical lines

	if (m_pImagePlane)
	{
		dbPainter.setPen(QPen(QColor(255,0,0), 2));

		const std::list<std::list<Vector2D<double> > *> &criticalLines = m_pImagePlane->getCriticalLineSegments();

		std::list<std::list<Vector2D<double> > * >::const_iterator it1;

		for (it1 = criticalLines.begin() ; it1 != criticalLines.end() ; it1++)
		{
			std::list<Vector2D<double> > *pLines = *it1;

			if (!pLines)
				continue;

			std::list<Vector2D<double> >::const_iterator it2 = pLines->begin();

			if (it2 == pLines->end())
				continue;

			Vector2D<double> p = *it2;
			int prevX, prevY;

			m_pImagePlane->getIndices(p, &prevX, &prevY);
			it2++;

			for ( ; it2 != pLines->end() ; it2++)
			{
				int curX, curY;

				p = *it2;
				m_pImagePlane->getIndices(p, &curX, &curY);

				dbPainter.drawLine(prevX+1, m_height-prevY-1, curX+1, m_height-curY-1);
				prevX = curX;
				prevY = curY;
			}

		}
	}
#endif
	// Actually draw everything on the screen
	pPainter->drawPixmap(0, 0, bufferPixmap);
}

void LensWidget::contentsMousePressEvent(QMouseEvent *pEvt)
{
	m_mouseStartX = pEvt->x();
	m_mouseStartY = pEvt->y();
	m_circleWidth = 0;
	m_lastDrawnWidth = 0;
	pEvt->accept();
}

void LensWidget::contentsMouseReleaseEvent(QMouseEvent *pEvt)
{
	if (m_mouseStartX >= 0 && m_mouseStartY >= 0 && m_circleWidth > 2)
	{
		if (m_circles.size() >= MAXCIRCLES)
		{
			QMessageBox::information(this, "Maximaal aantal lensonderdelen is bereikt",
					               "Maximaal aantal lensonderdelen is bereikt",
						       QMessageBox::Ok);


		}
		else
		{
			m_circles.push_back(LensCircle(m_mouseStartX, m_mouseStartY, m_circleWidth));
			
			// TODO: for testing
			applyLens();
		}
	}

	m_mouseStartX = -1;
	m_mouseStartY = -1;
	m_circleWidth = 0;
	m_lastDrawnWidth = 0;

	updateContents();

	pEvt->accept();
}

void LensWidget::contentsMouseMoveEvent(QMouseEvent *pEvt)
{
	int xDiff = pEvt->x() - m_mouseStartX;
	int yDiff = pEvt->y() - m_mouseStartY;
	
	int prevWidth = MAX(m_circleWidth, 0);

	m_circleWidth = MIN(ABS(xDiff), ABS(yDiff))*2;

	int startX = m_mouseStartX-m_circleWidth/2;
	int startY = m_mouseStartY-m_circleWidth/2;
	int startX2 = m_mouseStartX-m_lastDrawnWidth/2;
	int startY2 = m_mouseStartY-m_lastDrawnWidth/2;

	updateContents(startX-2, startY-2, m_circleWidth+6, m_circleWidth+6);
	updateContents(startX2-2, startY2-2, m_lastDrawnWidth+6, m_lastDrawnWidth+6);

	pEvt->accept();
}

class MyLensPlane : public LensPlane
{
public:
	MyLensPlane(const grale::GravitationalLens *pLens, Q3ProgressBar *pProgressBar) : LensPlane(pLens)
	{
		m_pProgressBar = pProgressBar;
		m_prevPct = -1;
	}

	void setFeedbackStatus(const std::string &msg)
	{
		std::cout << msg << std::endl;
	}

	void setFeedbackPercentage(int pct)
	{
		if (m_pProgressBar && pct != m_prevPct)
		{
			m_pProgressBar->setProgress(pct, 100);
			m_pProgressBar->repaint(false);
			m_prevPct = pct;
		}
	}
private:
	Q3ProgressBar *m_pProgressBar;
	int m_prevPct;
};

void LensWidget::applyLens()
{
	if (m_circles.empty()) // nothing to do
		return;

	CompositeLensParams compLensParams;
	CompositeLens compLens;

	double angularPixelSize = m_angularWidth/((double)m_width);
	double angularHeight = angularPixelSize*((double)m_height);
	double D_d = 1000.0*DIST_MPC;

	for (int i = 0 ; i < m_circles.size() ; i++)
	{
		double theta_E = angularPixelSize*(double)(m_circles[i].m_width)/2.0;
		double sigma_v = SQRT(theta_E*SPEED_C*SPEED_C/(4.0*CONST_PI));
		SISLensParams sisParams(sigma_v);

		SISLens sisLens;

		if (!sisLens.init(D_d, &sisParams))
			std::cerr << "Error initializing SIS lens: " << sisLens.getErrorString() << std::endl;
		else
		{
			double xPos = ((double)(m_circles[i].m_cx-1))*angularPixelSize-m_angularWidth/2.0;
			double yPos = ((double)(m_height-m_circles[i].m_cy))*angularPixelSize-angularHeight/2.0;

			if (!compLensParams.addLens(1, Vector2D<double>(xPos, yPos), 0, sisLens))
				std::cerr << "Error adding SIS lens to composite lens: " << compLensParams.getErrorString() << std::endl;
		}
	}

	if (!compLens.init(D_d, &compLensParams))
		std::cerr << "Error initializing composite lens: " << compLens.getErrorString() << std::endl;

	MyLensPlane lensPlane(&compLens,m_pProgressBar);
	Real2DLensPlaneFunction density(lensPlane);
	double criticalDensity = SPEED_C*SPEED_C/(4.0*CONST_PI*CONST_G)*1.0/D_d; // D_ds/D_s = 1 here

#ifdef SUBDIV
	if (!lensPlane.init(Vector2D<double>(-m_angularWidth,-angularHeight)/2.0,Vector2D<double>(m_angularWidth,angularHeight)/2.0, (m_width+1), (m_height+1), m_pCLKernel))
#else
	if (!lensPlane.init(Vector2D<double>(-m_angularWidth,-angularHeight)/2.0,Vector2D<double>(m_angularWidth,angularHeight)/2.0, m_width, m_height, m_pCLKernel))
#endif
	{
		std::cerr << "Error initializing lens plane: " << lensPlane.getErrorString() << std::endl;
		return;
	}
	m_pProgressBar->reset();

	m_massMap.resize(m_width*m_height);
	for (int y = 0 ; y < m_height ; y++)
		for (int x = 0 ; x < m_width ; x++)
#ifdef SUBDIV
			m_massMap[x+y*m_width] = density(Vector2D<double>(x,y)*angularPixelSize+Vector2D<double>(-m_angularWidth,-angularHeight)/2.0+Vector2D<double>(angularPixelSize,angularPixelSize)/2.0)/criticalDensity;
#else
			m_massMap[x+y*m_width] = density(Vector2D<double>(x,y)*angularPixelSize+Vector2D<double>(-m_angularWidth,-angularHeight)/2.0)/criticalDensity;
#endif

	SourcePlane *pSrcPlane = new SourcePlane(1.0,1.0);

	m_pImagePlane = new ImagePlane(&compLens, pSrcPlane);
	if (!m_pImagePlane->init(&lensPlane))
	{
		std::cerr << "Error initializing image plane: " << std::endl;
		return;
	}
	
	if (m_pCachedView)
	{
		delete m_pCachedView;
		m_pCachedView = 0;
	}
	updateContents();
}

void LensWidget::clearLens()
{
	if (m_pImagePlane == 0)
		return;

	m_circles.clear();
	delete m_pImagePlane->getSourcePlane();
	delete m_pImagePlane;
	m_pImagePlane = 0;
	m_massMap.resize(0);

	if (m_pCachedView)
	{
		delete m_pCachedView;
		m_pCachedView = 0;
	}

	updateContents();
}

void LensWidget::setMassScale(double s)
{
	m_massScale = s;

	if (m_massMap.empty())
		return;

	if (m_pCachedView)
	{
		delete m_pCachedView;
		m_pCachedView = 0;
	}
	updateContents();
}

void LensWidget::setLensBackground(QImage &image)
{
	if (image.width() != m_width || image.height() != m_height)
	{
		char str[1024];

		sprintf(str, "De resolutie van de achtergrond is %dx%d in plaats van %dx%d. De figuur zal herschaald worden.", 
			 image.width(),image.height(),m_width,m_height);
		QMessageBox::information(this, "Opmerking", str, QMessageBox::Ok);

		if (m_pBackGroundImage)
			delete m_pBackGroundImage;
		
		m_pBackGroundImage = new QImage();
		*m_pBackGroundImage = image.smoothScale(m_width, m_height, Qt::KeepAspectRatio);
	}
	else
	{
		if (m_pBackGroundImage)
			delete m_pBackGroundImage;
		m_pBackGroundImage = new QImage(image);
	}

	if (m_pCachedView)
	{
		delete m_pCachedView;
		m_pCachedView = 0;
	}
	updateContents();
}


QImage LensWidget::getLensedImage()
{
	QImage image(contentsWidth(), contentsHeight(), 32);
#ifdef SUBDIV
	int subDiv = 3;
	Vector2D<double> bottomLeft;
	double xStep = 0;
	double yStep = 0;

	if (m_pImagePlane)
	{
		bottomLeft = m_pImagePlane->getBottomLeft();
		xStep = m_pImagePlane->getXStep();
		yStep = m_pImagePlane->getYStep();
	}
		
	double xSubStep = xStep/(double)subDiv;
	double ySubStep = yStep/(double)subDiv;
	double div = subDiv*subDiv;

	m_pProgressBar->setProgress(0, 100);
	m_pProgressBar->repaint(false);

	int prevPct = 0;
#endif

	for (int y = 0 ; y < m_height ; y++)
	{
		for (int x = 0 ; x < m_width ; x++)
		{
			int R = 0;
			int G = 0;
			int B = 0;

			if (m_pBackGroundImage)
			{
#ifndef SUBDIV
				int x2 = x;
				int y2 = y;

				if (m_pImagePlane)
				{
					Vector2D<double> pos = m_pImagePlane->traceIndexCoordinate(x, y);
					m_pImagePlane->getIndices(pos, &x2, &y2);
				}

				y2 = m_height-y2-1;

				if (y2 >= 0 && x2 >= 0 && x2 < m_pBackGroundImage->width() && y2 < m_pBackGroundImage->height())
				{
					QRgb pixel = m_pBackGroundImage->pixel(x2, y2);

					R += qRed(pixel);
					G += qGreen(pixel);
					B += qBlue(pixel);
				}
#else
				if (m_pImagePlane)
				{
					double Rsum = 0;
					double Gsum = 0;
					double Bsum = 0;
					Vector2D<double> thetaStart = bottomLeft+Vector2D<double>(xStep*(double)x,yStep*(double)y)+Vector2D<double>(xSubStep,ySubStep)/2.0;

					for (int y2 = 0 ; y2 < subDiv ; y2++)
					{
						for (int x2 = 0 ; x2 < subDiv ; x2++)
						{
							Vector2D<double>theta = thetaStart+Vector2D<double>(xSubStep*(double)x2,ySubStep*(double)y2);
							Vector2D<double>beta;

							m_pImagePlane->traceThetaApprox(theta, &beta);

							Vector2D<double> diff = beta - bottomLeft;
							int x2 = diff.getX()/xStep;
							int y2 = diff.getY()/yStep;

							QRgb pixel = m_pBackGroundImage->pixel(x2, m_height-1-y2);

							Rsum += qRed(pixel);
							Gsum += qGreen(pixel);
							Bsum += qBlue(pixel);
						}
					}

					R += (int)(Rsum/div+0.5);
					G += (int)(Gsum/div+0.5);
					B += (int)(Bsum/div+0.5);
				}
				else
				{
					QRgb pixel = m_pBackGroundImage->pixel(x, m_height-1-y);

					R += qRed(pixel);
					G += qGreen(pixel);
					B += qBlue(pixel);
				}
#endif
			}

			if (!m_massMap.empty())
			{
				double density = m_massMap[x+y*m_width]*m_massScale;
				double density2 = MIN(1.0,density);
				
				density2 = MAX(0.0,density2);

				double y = SIN(density2)/SIN(1.0);
				double x = y*y;

				int Rvalue = (int)(255.0*x);
				int Gvalue = (int)(192.0*x);

				R += Rvalue;
				G += Gvalue;
			}

			R = MIN(255, R);
			G = MIN(255, G);
			B = MIN(255, B);

			image.setPixel(x, m_height-1-y, qRgb(R,G,B));
		}

#ifdef SUBDIV
		int pct = (int)(((double)(y+1)/(double)(m_height))*100.0+0.5);
		if (prevPct != pct)
		{
			prevPct = pct;
			m_pProgressBar->setProgress(pct, 100);
			m_pProgressBar->repaint(false);
		}
#endif
	}
#ifdef SUBDIV
	m_pProgressBar->reset();
#endif

	return image;
}

