#ifndef LENSWIDGET_H

#define LENSWIDGET_H

#include <q3scrollview.h>
#include <q3progressbar.h>
//Added by qt3to4:
#include <QMouseEvent>
#include <vector>

namespace grale
{
	class ImagePlane;
	class OpenCLKernel;
}

class QImage;
class Q3ProgressBar;

class LensWidget : public Q3ScrollView
{
	Q_OBJECT;
public:
	LensWidget(QWidget *pParent, int width, int height, double angularWidth);
	~LensWidget();

	void setProgressBar(Q3ProgressBar *pProgressBar) 						{ m_pProgressBar = pProgressBar; }
	void setLensBackground(QImage &image);
	QImage getLensedImage();
public slots:
	void clearLens();
	void setMassScale(double s);
private:
	void drawContents(QPainter *pPainter, int clipx, int clipy, int clipw, int cliph);
	void contentsMousePressEvent(QMouseEvent *pEvt);
	void contentsMouseReleaseEvent(QMouseEvent *pEvt);
	void contentsMouseMoveEvent(QMouseEvent *pEvt);

	void applyLens();

	Q3ProgressBar *m_pProgressBar;

	int m_width, m_height;
	double m_angularWidth;

	int m_mouseStartX, m_mouseStartY;
	int m_circleWidth, m_lastDrawnWidth;

	class LensCircle
	{
	public:
		LensCircle(int cx, int cy, int w) { m_cx = cx; m_cy = cy ; m_width = w; }
		
		int m_cx, m_cy, m_width;
	};

	std::vector<LensCircle> m_circles;

	grale::ImagePlane *m_pImagePlane;
	std::vector<double> m_massMap; // in units of the critical density
	QImage *m_pBackGroundImage;
	QImage *m_pCachedView;

	double m_massScale;

	grale::OpenCLKernel *m_pCLKernel;
};

#endif // LENSWIDGET_H
